﻿/*
 * Written by Seth A. Robinson (rtsoft.com)
 * 
*/

using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.Networking;
using System.Runtime.Serialization.Formatters.Binary;
using System.IO;
using System;

public class RTNetworkClient : MonoBehaviour, IRTNetworkConnection
{
    public event Action OnConnectedEvent;
    public event Action OnConnectFailedEvent;
    public event Action OnDisconnectedEvent;
    public event Action<string> OnPacketStringEvent;
    public event Action<float> OnPacketPingEvent;

    //void OnPacketEvent(int packetType, byte[] buffer, int size, int index);
    public event Action<short, byte[], int, int> OnPacketEvent;

    public bool isConnected { get; private set; }

    public int GetHostID() { return _hostID; }
    public int GetConnectionID() { return _connectionID; }
    
    int _connectionID;
    int _hostID = -1;
    static byte[] _buffer;
    const int _bufferSize = 2048;

    string _connectASAPDestinationSite;
    int _connectASAPport;
    bool _shouldShowDebugMessages = false;
    float _timeOfLastPingSend = 0;


   void Start()
    {
        _buffer = new byte[_bufferSize];
    }

  
    public void SetShowDebugMessages(bool bNew) { _shouldShowDebugMessages = bNew; }

    void OnDestroy()
    {
        Disconnect();
    }

    //note: This is NOT a polite way to disconnect, we kill the whole port right after sending the request so I doubt it even gets sent.  However, your server DOES
    //need to handle this kind of disconnect so I didn't bother with a polite way for now

    public void Disconnect()
    {
        if (_hostID != -1)
        {
            if (_connectionID > 0)
            {

                Log("Disconnecting hostID "+_hostID+" client connectionID " + _connectionID);

                byte error;
                NetworkTransport.Disconnect(_hostID, _connectionID, out error);

                if ((NetworkError)error != NetworkError.Ok)
                {
                    Log("Error disconnecting connectionid "+_connectionID+" - Error " + (NetworkError)error);
                    OnConnectFailedEvent();
                }
            }
        
            NetworkTransport.RemoveHost(_hostID);
            _hostID = -1;
            _connectionID = 0;
        } else
        {
           // Log("Not removing host, it's not initted");
        }

        //either way, init the whole thing again, otherwise we start getting "wronghost" errors after 12 connects

       // RTNetworkManager.Get().InitNetworkStuff(2);
    }

    void Log(string text)
    {
        if (_shouldShowDebugMessages) RTConsole.Log(text);
    }
    
    public bool Connect(string destinationSite, int port)
    {
        _connectASAPDestinationSite = "";
        _connectASAPport = 0;
        byte error;

        if (_hostID != -1)
        {
            Log("Client disconnecting before reconnecting... host "+_hostID);
            //well, we should disconnect first
            Disconnect();
        }

        _hostID = NetworkTransport.AddHost(RTNetworkManager.Get().topology, 0); //0 means we don't care

        if (_hostID == -1)
        {
            Log("Error connecting to " + destinationSite + " at " + port + " - Error hostID is -1, reached maximum hosts probably");
            OnConnectFailedEvent();
            return false;
        }

        string ipAddress = "error";


        #if !UNITY_WEBGL
       
        System.Net.IPAddress[] iplist = System.Net.Dns.GetHostAddresses(destinationSite);

        if (iplist == null || iplist.Length == 0)
        {
            
            Log("Can't convert "+iplist+" to an IP address!");
            OnConnectFailedEvent();
            return false;
        }

        foreach (System.Net.IPAddress address in iplist)
        {
            if (address.AddressFamily == System.Net.Sockets.AddressFamily.InterNetwork)
            {
                //looks legit
                ipAddress = address.ToString();
            }
        }
#else
        //webGL coughs and dies on System.Net.Dns.GetHostAddresses so I hope you used the actual numeric IP address, fool
        
        if (destinationSite == "localhost")
        {
            destinationSite = "127.0.0.1";
        }
        ipAddress = destinationSite;

        
#endif
        _connectionID = NetworkTransport.Connect(_hostID, ipAddress, port, 0, out error);

        Log("Connecting to " + destinationSite + " ("+ipAddress+") :" + port + " with hostID "+_hostID+".  Was given connection id: " + _connectionID + ") ...");

        if ((NetworkError)error  != NetworkError.Ok)
        {
           Log("Error connecting to " + destinationSite + " at " + port+" - Error "+ (NetworkError)error);
           OnConnectFailedEvent();
            return false;
        }

        return true; //success
    }
    

    public bool IsWaitingForPing()
    {
        return _timeOfLastPingSend != 0;
    }

    public void Ping()
    {
        if (_hostID < 0 || _connectionID <= 0) return;
        _timeOfLastPingSend = Time.time;
        RTNetworkManager.Get().SendPacketNoData(_hostID, _connectionID, (short)RTNetworkManager.PacketType.PING);
    }

    // Update is called once per frame
    void Update()
    {

        if (_hostID == -1) return;

        int outConnectionId;
        int outChannelId;
        int receiveSize;
        byte error;
        NetworkEventType evnt;

        short packetType = 0;
        int index = 0;

        while ( (evnt = NetworkTransport.ReceiveFromHost(_hostID, out outConnectionId, out outChannelId, _buffer, _bufferSize, out receiveSize, out error)) != NetworkEventType.Nothing)
        {

            if ((NetworkError)error != NetworkError.Ok)
            {
                Log("Client error getting packet from " + _hostID + " - Error: " + (NetworkError)error);

                //disconnect I guess.  This is possibly the wrong way to do it but...
                NetworkTransport.RemoveHost(_hostID);
                _hostID = -1;
                _connectionID = 0;

                if (OnDisconnectedEvent != null) OnDisconnectedEvent();
                break;
            }

            if (_connectionID != outConnectionId)
            {
                Log("Ignoring packet from old connection ID " + outConnectionId);
                continue;
            }

            if (_shouldShowDebugMessages)
            {
                //Log("Client: Got " + evnt + " outConnectionId: " + outConnectionId + ", size: " + receiveSize+" channel: "+outChannelId);
            }

                switch (evnt)
                {

                    case NetworkEventType.DataEvent:
                    if (receiveSize < 2)
                    {
                        Log("Client: Bad data packet, only " + receiveSize + " bytes");
                        return;
                    }

                    index = 0;
                    RTUtil.SerializeInt16(ref packetType, _buffer, ref index);
                    
                    switch (packetType)
                    {
                        case (short)RTNetworkManager.PacketType.STRING:

                            string s = "";
                            RTUtil.SerializeString(ref s, _buffer, ref index);
                            OnPacketStringEvent(s);
                            break;


                        case (short)RTNetworkManager.PacketType.PING_RECEIVED:
                            float pingTime = Time.time - _timeOfLastPingSend;
                            _timeOfLastPingSend = 0;
                            OnPacketPingEvent(pingTime);
                            break;

                        default:
                            OnPacketEvent(packetType, _buffer, receiveSize, index);
                            break;

                    }

                    break;

                    case NetworkEventType.ConnectEvent:
                        isConnected = true;

                    if (OnConnectedEvent != null) OnConnectedEvent();
                        break;

                    case NetworkEventType.DisconnectEvent:

                        if (isConnected)
                        {
                            isConnected = false;

                            _connectionID = 0;

                            if (_connectASAPDestinationSite.Length > 0)
                            {
                                //we're about to connect again, don't bother with this event, it would be confused with the reconnect
                                Connect(_connectASAPDestinationSite, _connectASAPport);
                            }
                            else
                            {

                              if (OnDisconnectedEvent != null) OnDisconnectedEvent();
                            }
                        }
                        else
                        {
                            OnConnectFailedEvent();
                        }
                        break;
                }
          
        }
               
    }
}
