﻿/*
 * Written by Seth A. Robinson (rtsoft.com)
 * 
*/
using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.Networking;
using System.Runtime.Serialization.Formatters.Binary;
using System.IO;

public class RTNetworkManager : MonoBehaviour
{
    public static int myReliableChannelId { get; private set; }
    public static int myUnreliableChannelId { get; private set; }

    public enum PacketType
    {
        PING,
        PING_RECEIVED,
        STRING,
        DB,
        RESERVED = 100 //add your own type starting at 101
    }
    
    ConnectionConfig config;
    static RTNetworkManager _this;
    public HostTopology topology;
    public static RTNetworkManager Get() { return _this; }
    byte[] _buffer;

    void Awake()
    {
        _buffer = new byte[2048];
        DontDestroyOnLoad(gameObject);
        _this = this;
    }

    // Use this for initialization
    public void InitNetworkStuff (int maxConnections, ushort maxHosts, bool setupAsServer)
    {

        // Initializing the Transport Layer with no arguments (default settings)
         GlobalConfig gConfig = new GlobalConfig();

        //gConfig.MaxPacketSize = 500; //Don't use custom settings unless you have to, says Unity

        if (setupAsServer)
        {

            //if you get "Error: Cannot allocate new packet for sending" suring burst sends you need to raise these.  The docs say they auto increase but it seems like they only do AFTER hitting the errors
            gConfig.ReactorMaximumSentMessages = 30000;
            gConfig.ReactorMaximumReceivedMessages = 30000;
            gConfig.ReactorModel = ReactorModel.SelectReactor;
        }
      
        gConfig.ThreadAwakeTimeout = 1;

        gConfig.MaxHosts = maxHosts; //helps with webgl as re-using hosts causes webgl freezes for some reason!?
        
        NetworkTransport.Init(gConfig);

        config = new ConnectionConfig();

        config.AcksType = ConnectionAcksType.Acks32; //NOTE: Must match server or you'll get a CRC error on connection
        config.PingTimeout = 4000; //NOTE: Must match server or you'll get a CRC error on connection
        config.DisconnectTimeout = 20000; //NOTE: Must match server or you'll get a CRC error on connection

        if (setupAsServer)
        {
            config.MaxSentMessageQueueSize = 30000; //Um... is this per connection or?
        }

        config.MaxCombinedReliableMessageCount = 1;
        config.MaxCombinedReliableMessageSize = 100; //in bytes
        config.MinUpdateTimeout = 1;
        config.OverflowDropThreshold = 100;
        config.NetworkDropThreshold = 100;
        config.SendDelay = 0;
        
        myReliableChannelId = config.AddChannel(QosType.Reliable);
        myUnreliableChannelId = config.AddChannel(QosType.Unreliable);
        topology = new HostTopology(config, maxConnections);

        if (setupAsServer)
            topology.SentMessagePoolSize = ushort.MaxValue; //this # is very important when bursting packets out

        /*
        RTConsole.Log("Bandwidth allowed: " + config.InitialBandwidth + " Send queue max: " + config.MaxSentMessageQueueSize + " Max hosts: " + maxHosts + " max connections: " + maxConnections+" Sent message pool size: " + topology.SentMessagePoolSize+ " ReactorMax rec: " + gConfig.ReactorMaximumReceivedMessages+
            " Max send pool size: "+ topology.SentMessagePoolSize);
            */

       
    }

    public string GetStats()
    {

        string s, b, z;

#if UNITY_WEBGL
        s = "(webgl doesn't support all stats)";
        b = "";
        z = "";
#else

        byte error;

        /*
        These give the error:  the function called has not been supported for web sockets communication

        int webQueueIn = NetworkTransport.GetIncomingMessageQueueSize(RTNetworkServer.Get().GetWebHostID(), out error);
        if ((NetworkError)error != NetworkError.Ok)
        {
            RTConsole.Log("NetworkTransport.GetIncomingMessageQueueSize - Error " + (NetworkError)error);
        }

        int webQueueOut = NetworkTransport.GetOutgoingMessageQueueSize(RTNetworkServer.Get().GetWebHostID(), out error);
        if ((NetworkError)error != NetworkError.Ok)
        {
            RTConsole.Log("NetworkTransport.GetOutgoingMessageQueueSize - Error " + (NetworkError)error);
        }

         int webBytesOut = NetworkTransport.GetOutgoingFullBytesCountForHost(RTNetworkServer.Get().GetWebHostID(), out error);
        if ((NetworkError)error != NetworkError.Ok)
        {
            RTConsole.Log("GetOutgoingFullBytesCountForHost - Error " + (NetworkError)error);
        }
    */





        int userPayloadBytes = NetworkTransport.GetOutgoingUserBytesCount();
        int totalBytesSent = NetworkTransport.GetOutgoingFullBytesCount();
        int systemBytesSent = NetworkTransport.GetOutgoingSystemBytesCount();
        int UNETOverhead = systemBytesSent-userPayloadBytes; //This seems to be right.  GetOutgoingSystemBytesCount() also includes the user paypload, so let's get rid of that

        float overhead;
        if (totalBytesSent != 0)
        {
            overhead = (float)UNETOverhead / (float)totalBytesSent;
        }
        else
        {
            overhead = 0;
        }

        s = "Network timestamp: " + NetworkTransport.GetNetworkTimestamp()+"\r\n";
        s += "WebGL: Unity doesn't give stats for WebGL, so this is all UDP only below";
       b = "UDP: In Msg Queue: " + NetworkTransport.GetIncomingMessageQueueSize(RTNetworkServer.Get().GetHostID(), out error)
            + " Out queue: " + NetworkTransport.GetOutgoingMessageQueueSize(RTNetworkServer.Get().GetHostID(), out error);
        z = "Total packets sent: " + NetworkTransport.GetOutgoingPacketCount() + " Total bytes sent: " + NetworkTransport.GetOutgoingFullBytesCount()
            + " (" + ((totalBytesSent / 1024.0f) / 1024.0f).ToString("0.#") + " MB)" +
            " Unity packet overhead: " + (overhead * 100).ToString("0.##") + "%. (" + UNETOverhead + " bytes) Payload bytes: " + userPayloadBytes + " System (system+payload?): " + systemBytesSent
        + " Packets rec: "+NetworkTransport.GetIncomingPacketCountForAllHosts();

        z += "\r\nPackets dropped: " + NetworkTransport.GetIncomingPacketDropCountForAllHosts();
            

#endif


        return s + "\r\n" + b + "\r\n" + z;

    }


    // Update is called once per frame
    void Update ()
    {
		
	}

    public void SendString(IRTNetworkConnection client, string message)
    {
        byte error;
        
        int index = 0;
        short packetType = (short)PacketType.STRING;
        RTUtil.SerializeInt16(packetType, _buffer, ref index);
        RTUtil.SerializeString(message, _buffer, ref index);

        NetworkTransport.Send(client.GetHostID(), client.GetConnectionID(), myReliableChannelId, _buffer, index, out error);

        if ((NetworkError)error != NetworkError.Ok)
        {
            RTConsole.Log("SendString: Error sending to host " + client.GetHostID() + " ConID: "+client.GetConnectionID()+" : - Error " + (NetworkError)error);
        }
    }

    public void SendPacketNoData(int tempHostID, int tempConnectionId, short packetType)
    {
        byte error;
        int index = 0;
        RTUtil.SerializeInt16(packetType, _buffer, ref index);
        NetworkTransport.Send(tempHostID, tempConnectionId, myReliableChannelId, _buffer, index, out error);

        if ((NetworkError)error != NetworkError.Ok)
        {
            RTConsole.Log("SendPacketNoData: Error sending to host " + tempHostID + " ConID: " + tempConnectionId + " : - Error " + (NetworkError)error);
        }
    }

}
