﻿/*
 * Written by Seth A. Robinson (rtsoft.com)
 * 
*/
using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.Networking;
using System.Runtime.Serialization.Formatters.Binary;
using System.IO;
using System;

public interface IRTNetworkConnection
{
    int GetHostID();
    int GetConnectionID();
}

public class RTClientConnection : IRTNetworkConnection
{
    public int GetHostID() { return hostID; }
    public int GetConnectionID() { return connectionID; }

    public int hostID;
    public int connectionID;
    public string ip;
    public UInt32 key; //unique key for this connection, same one we are indexed in the dictionary with
    
}

public class RTNetworkServer : MonoBehaviour
{

   // List<RTClientConnection> clients;

    Dictionary<UInt32, RTClientConnection> _clients;

    int _hostID = -1;
    int _webHostID = -1;
    public static RTNetworkServer _this;

    public event Action<string, RTClientConnection> OnPacketStringEvent;
    
    public static RTNetworkServer Get() { return _this; }
    byte[] _buffer;
    const int _bufferSize = 2048;
    bool _shouldShowDebugMessages;
    bool _didSetup;

    //used for stat keeping
    float _timeOfLastStats;
    float _timeOfLastFrameStart;
    float _biggestFrameLagSinceLastStats;
    int _framesSinceLastStats;
    float _timeOfLastFrame;
    int _packetsSentAtLastStats;
    int _bytesSentAtLastStats;
    int _packetsRecSentAtLastStats;


    void Awake ()
    {
        _this = this;
        DontDestroyOnLoad(gameObject);
        _clients = new Dictionary<uint, RTClientConnection>();
    }

    public void SetShowDebugMessages(bool bNew) { _shouldShowDebugMessages = bNew; }
    void Log(string text)
    {
        RTConsole.Log(text);
    }

    public int GetHostID() { return _hostID; }
    public int GetWebHostID() { return _webHostID; }

    public bool DidSetup() { return _didSetup; }
    
    public bool Setup(int port, int webGLPort)
    {

        if (_didSetup)
        {

            Log("Server can't be initted twice! Ignoring Setup() call");
            return false;
        }
       
        _hostID = NetworkTransport.AddHost(RTNetworkManager.Get().topology, port);
        
        if (_hostID == -1)
        {
            Log("Error initializing UDP port " + port + ", it's probably already being used.");
            return false;
        }

        _webHostID = NetworkTransport.AddWebsocketHost(RTNetworkManager.Get().topology, webGLPort);
        
        if (_webHostID == -1)
        {
            Log("Error initializing webGL port " + port + ", it's probably already being used.");
            return false;
        }

        _didSetup = true;
        _buffer = new byte[_bufferSize];
        Log("Initted UDP server port at " + port + ", host id is " + _hostID+", WebGL server at port "+webGLPort+" host id is "+_webHostID);

        ResetFrameStats();
        return true;
    }

    void ResetFrameStats()
    {
        _timeOfLastStats = Time.time;
        _timeOfLastFrameStart = 0;
        _biggestFrameLagSinceLastStats = 0;
        _framesSinceLastStats = 0;

        #if !UNITY_WEBGL
        _packetsSentAtLastStats = NetworkTransport.GetOutgoingPacketCount();
        _bytesSentAtLastStats = NetworkTransport.GetOutgoingFullBytesCount();
        _packetsRecSentAtLastStats = NetworkTransport.GetIncomingPacketCountForAllHosts();
        #endif

    }

    // Update is called once per frame

    void FixedUpdate()
    {
       // print("Fixed Update tick " + Time.time + " : " + Time.fixedDeltaTime);

    }

    public void SendStringToAll(string message)
    {
        foreach(RTClientConnection client in _clients.Values)
        {
            RTNetworkManager.Get().SendString(client, message);
        }
    }
    
    void HandleEvents(int tempHostID)
    {
        NetworkEventType evnt;

        int outConnectionId;
        int outChannelId;
        int receiveSize;
        byte error;
        UInt32 key;

        while ((evnt = NetworkTransport.ReceiveFromHost(tempHostID, out outConnectionId, out outChannelId, _buffer, _bufferSize, out receiveSize, out error)) != NetworkEventType.Nothing)
        {
            if (_shouldShowDebugMessages)
                Log("Server host "+ tempHostID+": Got " + evnt + " from connectionID " + outConnectionId + ", channel " + outChannelId + " size " + receiveSize+" "+_clients.Count+" clients connected");

            key = (UInt32)(UInt16)tempHostID << 16 | (UInt16)outConnectionId;

            switch (evnt)
            {
                case NetworkEventType.DataEvent:

                    if (outChannelId == RTNetworkManager.myReliableChannelId ||
                        outChannelId == RTNetworkManager.myUnreliableChannelId)
                    {

                        if (receiveSize < 2)
                        {
                            if (_shouldShowDebugMessages)
                                Log("Server: Bad data packet size of " + receiveSize);
                            continue;
                        }

                        short packetType = 0;
                        int index = 0;
                        RTUtil.SerializeInt16(ref packetType, _buffer, ref index);

                        switch (packetType)
                        {
                            case (short)RTNetworkManager.PacketType.STRING:
                                string s = "";
                                RTUtil.SerializeString(ref s, _buffer, ref index);
                                OnPacketStringEvent(s, _clients[key]);
                                break;

                            case (short)RTNetworkManager.PacketType.PING:
                                RTNetworkManager.Get().SendPacketNoData(tempHostID, outConnectionId, (short)RTNetworkManager.PacketType.PING_RECEIVED);
                                break;

                            default:
                                //OnPacketEvent(packetType, _buffer, receiveSize, index);
                                break;
                        }
                        
                    } else
                    {
                        //um.. webgl sends weird 3 byte packets on channel 110.  Some kind of keep alive?
                    }
                    break;

                case NetworkEventType.ConnectEvent:

                    if (outChannelId > 64000)
                    {
                        Debug.LogError("ConnectionID of > 64000 found.  That should be impossible.. this may break the key hashing!");
                    }

                    RTClientConnection client = new RTClientConnection();
                    client.hostID = tempHostID;
                    client.connectionID = outConnectionId;
                    client.key = key;
                    int clientPort;
                    UnityEngine.Networking.Types.NetworkID clientNetID;
                    UnityEngine.Networking.Types.NodeID clientNodeID;

                    if (tempHostID != _webHostID)
                    {
                        //this gets a Unity error if you try to use it for a WebSocket connection, so UDP only I guess
                        NetworkTransport.GetConnectionInfo(tempHostID, outConnectionId, out client.ip, out clientPort, out clientNetID, out clientNodeID, out error);
                        if ((NetworkError)error != NetworkError.Ok)
                        {
                            Log("NetworkTransport.GetConnectionInfo " + tempHostID + " - Error: " + (NetworkError)error);
                        }

                    }
                        _clients.Add(key, client);
                    //if (_shouldShowDebugMessages)
                        Log("Host " + tempHostID + " added a new client from "+client.ip+" with connectionID " + outConnectionId + ". " + _clients.Count + " clients online");
                    break;

                case NetworkEventType.DisconnectEvent:
                    
                    //if (_shouldShowDebugMessages)
                        Log("Host " + tempHostID + " removed client "+ _clients[key].ip+" connectionID " + outConnectionId + ". " + _clients.Count + " clients online");

                    _clients.Remove(key);
                     break;

            }

        }
    }

    public string GetStats()
    {
        string s;

        float timeSinceLastStats = Time.time- _timeOfLastStats;
        #if !UNITY_WEBGL

        int packetsSent = NetworkTransport.GetOutgoingPacketCount() - _packetsSentAtLastStats;
        int bytesSent = NetworkTransport.GetOutgoingFullBytesCount() - _bytesSentAtLastStats;
        int packetsRec = NetworkTransport.GetIncomingPacketCountForAllHosts() - _packetsRecSentAtLastStats;
        #endif

        string fps = "?";

        if (timeSinceLastStats > 0)
        {
            fps = ((float)_framesSinceLastStats/ timeSinceLastStats).ToString("0.####");
        }

        s = "Clients: " + _clients.Count+" ";
          s += "Stats from last " + timeSinceLastStats.ToString("0.##") + " seconds - Average FPS:" + fps + " FPS)\r\nTime of last frame: "
          + _timeOfLastFrame.ToString("0.####") + " Slowest frame " + _biggestFrameLagSinceLastStats.ToString("0.####");

#if UNITY_WEBGL
           s += " Packets sent: Data not available on WebL build";
#else
          s += " Packets sent: " + packetsSent + " Bytes sent: " + bytesSent + " ("+( ((float)bytesSent/1024.0f)/1024.0f).ToString("0.#")+" MB) Packets Received: " + packetsRec;
#endif

        ResetFrameStats();
        return s;
    }
    void Update()
    {
        if (!_didSetup || _webHostID == -1 || _hostID == -1) return;

        // print("Update tick A " + Time.time + " : " + Time.deltaTime);

        if (_timeOfLastFrameStart != 0)
        {
            //we have valid data, compute stat stuff
            _timeOfLastFrame = Time.time - _timeOfLastFrameStart;

            _framesSinceLastStats++;
            if (_timeOfLastFrame > _biggestFrameLagSinceLastStats)
            {
                _biggestFrameLagSinceLastStats = _timeOfLastFrame;
            }

        }

        HandleEvents(_hostID);
        HandleEvents(_webHostID);

        _timeOfLastFrameStart = Time.time;
    }
}
