/*

Source code by Seth A. Robinson

 */

//#define RT_NOAUDIO

using UnityEngine;
using System.Collections;
using System;
using System.IO;
using System.Text;
using UnityEngine.UI;
using System.Collections.Generic;

public class GameLogic : MonoBehaviour
{

    public int serverUDPPort = 45345;
    public int serverWebGLPort = 45346;
    public int serverTelnetPort = 45347;
    bool showExtraDebugLogging = false; //change to true to see more stuff

    int maxConnectionsAtOnce = 3; //gets overriden if in test or server mode later
    public string serverAddress = "localhost"; //Remote server IP to connect to.  WARNING: You need to put the actual IP and not the server name, otherwise the WebGL version will
                                           //fail because System.Net.Dns.GetHostAddresses doesn't work.   gets overriden to localhost if "Host" button is clicked.
    bool isHeadlessServer = false; //gets overriden later if -batchmode parm is detected
    bool isDebugClient = true; //forces server settings even on the client, causes extra hosts and packet maxes to be initialized, wastes a lot of memory/resources but better for stress testing and the like

    static GameLogic _this = null;

    GameClient _gameClient;
    GameServer _gameServer;

    public static string GetName()
    {
        return Get().name;
    }

    private void Awake()
    {
        _gameClient = GetComponent<GameClient>();
        _gameServer = GetComponent<GameServer>();
        
        Application.targetFrameRate = 20000;
        QualitySettings.vSyncCount = 0;
        
        Debug.unityLogger.filterLogType = LogType.Log;
        Application.SetStackTraceLogType(LogType.Log, StackTraceLogType.ScriptOnly);
        Application.SetStackTraceLogType(LogType.Error, StackTraceLogType.ScriptOnly);
        Application.SetStackTraceLogType(LogType.Assert, StackTraceLogType.ScriptOnly);
        Application.SetStackTraceLogType(LogType.Exception, StackTraceLogType.ScriptOnly);
        Application.SetStackTraceLogType(LogType.Warning, StackTraceLogType.ScriptOnly);
        
    }
    // Use this for initialization
    void Start()
    {
        // RTAudioManager.Get().SetDefaultMusicVol(0.4f);
        _this = this;
        _gameServer.SetPorts(serverUDPPort, serverWebGLPort, serverTelnetPort);

#if RT_NOAUDIO
		AudioListener.pause = true;
#endif
  
        _gameClient.SetShowDebugMessages(showExtraDebugLogging);
        RTNetworkServer.Get().SetShowDebugMessages(showExtraDebugLogging);


        string commandLineOptions = System.Environment.CommandLine;
        if (commandLineOptions.Contains("-batchmode") || commandLineOptions.Contains("-servermode") || isHeadlessServer)
        {
            isDebugClient = false;
            RTConsole.Get().SetHeadlessMode(true);
            isHeadlessServer = true;
            RTConsole.Get().SetMirrorToDebugLog(true);
            RTConsole.Get().SetShowUnityDebugLogInConsole(false);
            GameObject.Find("RTConsole").SetActive(false); //don't waste rendering power, we won't use the display part
            Destroy(GameObject.Find("DebugCanvas"));
            Destroy(GameObject.Find("FramerateCounter"));

            /*
            string[] args = Environment.GetCommandLineArgs();
            foreach (string arg in args)
            {
                RTConsole.Log("Arg detected: " + arg);
            }
            */
        }
        else
        {
            RTConsole.Get().SetShowUnityDebugLogInConsole(true);
        }


#if UNITY_EDITOR
        // Debug.Log("Unity Editor detected, setting debug mode on "+ serverAddress);
#endif
        //RTEventManager.Get().Schedule(RTAudioManager.GetName(), "PlayMusic", 1, "intro");
      
        ushort maxHosts = 3;
     
        if (IsHeadlessServer())
        {

             //extra super strong settings
            maxConnectionsAtOnce = 1000;
            Application.targetFrameRate = 20000;
            maxHosts = 128; //not really needed, but whatever
            QualitySettings.vSyncCount = 0;
            //Time.fixedDeltaTime = 0.02f;
            //Time.maximumDeltaTime = 0.02f;
          
            RTNetworkManager.Get().InitNetworkStuff(maxConnectionsAtOnce, maxHosts, true);
            RTConsole.Log("`8Initialized `4NetTest`` in headless server mode -  Target framerate: " + Application.targetFrameRate + ". Fixed delta: " + Time.fixedDeltaTime);
            _gameServer.InitServer();
        }
        else
        {

            if (isDebugClient)
            {
                //extra super strong settings so stress testing works
                maxConnectionsAtOnce = 1000;
                maxHosts = 128; 
            }
            //not a server, but still need to init this stuff
            RTNetworkManager.Get().InitNetworkStuff(maxConnectionsAtOnce, maxHosts, isDebugClient);
            //DoClientConnect();
            RTConsole.Log("NetTest `wV1.00`` by Seth A. Robinson (`wrtsoft.com``).");
            RTConsole.Log("Server settings: `w"+ serverAddress + "`` UDP port: `w"+serverUDPPort+"`` WebSocket port: `w"+serverWebGLPort+"`` Telnet port: `w"+serverTelnetPort+"``");
            RTConsole.Log("Click Host, then UDP or WebGL connect to connect to yourself.  Or, if a remote server exists, just click UDP or WebGL connect to connect to it.  (Change the IP/port by editing the top of `wGameLogic.cs``)");

        }

        if (!isDebugClient)
        {
            Destroy(GameObject.Find("Stress Test Debug Panel"));
            Destroy(GameObject.Find("Server Debug Panel"));

        }

    }

    public void DoClientConnect()
    {
        _gameClient.Connect(serverAddress, GetCorrectServerPort());
    }

    public string GetServerAddress()
    {
        return serverAddress;
    }

    public int GetCorrectServerPort()
    {
       if (Application.platform == RuntimePlatform.WebGLPlayer)
       {
          return serverWebGLPort;
       }

        return serverUDPPort;
    }

    public bool IsHeadlessServer()
    {
        return isHeadlessServer;
    }

   public void SetServerURL(string url)
    {
        serverAddress = url;
    }

    public void ConnectUDP()
    {
       _gameClient.Connect(serverAddress, serverUDPPort);
    }

    public void ConnectWebGL()
    {
        _gameClient.Connect(serverAddress, serverWebGLPort);
    }

    
    static public GameLogic Get()
	{
		return _this;
	}
 
	void OnApplicationQuit() 
	{
        // Make sure prefs are saved before quitting.
        //PlayerPrefs.Save();
        RTConsole.Log("Application quitting normally");
        print("QUITTING!");
    }

    public void ShowStats()
    {
        if (_gameClient.GetNetworkClient().isConnected)
        {
            RTNetworkManager.Get().SendString(_gameClient.GetNetworkClient(), "/stats");
        } else
        {
            RTConsole.Log("Main client is not connected to a server");
        }
        
    }
    
	public void StartGame()
	{
		//RTAudioManager.Get().Play("music_box");
	}

	// Update is called once per frame
	void Update () 
    {

    }

}
